#include "scm.h"

Command commands[0x5C3] = {
#include "com.inc"
};

void
scriptInit(Script *sc)
{
	uint nextblock;
	ushort numNames;
	uchar *buf;
	int i;

	nextblock = *(uint*)(sc->buf+3);

	buf = sc->buf + nextblock;
	nextblock = *(uint*)(buf+3);
	numNames = *(ushort*)(buf+8);
	sc->numNames = numNames;
	sc->names = (char(*)[24])(buf+12);

	buf = sc->buf + nextblock;
	nextblock = *(uint*)(buf+3);
	sc->mainOffset = nextblock;
	sc->mainEnd = *(int*)(buf+8);
	sc->largest = *(int*)(buf+12);
	sc->numMissions = *(short*)(buf+16);
	sc->numExclusive = *(short*)(buf+18);
	sc->missionOffsets = (int*)(buf+20);
}

void
disassemble(FILE *out, Script *sc, uchar *miss, uchar *end)
{
	int i, j;
	ushort cmd;
	uchar type;

	while(miss < end){
		cmd = *(ushort*)miss;
		miss += 2;
		fprintf(out, "%X %X\n", miss-sc->buf-2, cmd);
		if(cmd & 0x8000){
			fprintf(out, "cond ");
			cmd &= ~0x8000;
		}
//		fprintf(out, "%s ", commands[cmd].name);
		for(i = 0; i < 18 && commands[cmd].args[i]; i++){
			type = *miss++;
//		fprintf(out, "\n\t%d %X %X ", i, commands[cmd].args[i], type);
			switch(type){
			case 0:
				i = 42;	// break from loop
				break;
			case Long:
				fprintf(out, "%d ", *(int*)miss);
				miss += 4;
				break;
			case Local:
				fprintf(out, "local_%x ", *(short*)miss);
				miss += 2;
				break;
			case Global:
				fprintf(out, "global_%x ", *(short*)miss);
				miss += 2;
				break;
			case Word:
				fprintf(out, "%d ", *(short*)miss);
				miss += 2;
				break;
			case Byte:
				fprintf(out, "%d ", *miss++);
				break;
			case Float:
				fprintf(out, "%f ", *(short*)miss / 16.0f);
				miss += 2;
				break;
			default:
				fprintf(out, "\"%s\" ", miss-1);
				miss += 7;
				break;
			}
		}
		fprintf(out, "\n");
	}
}

void
dumpScriptInfo(Script *sc)
{
	int i;

	printf("numNames: %d\n", sc->numNames);
	printf("mainEnd: 0x%x\n", sc->mainEnd);
	printf("largest: 0x%x\n", sc->largest);
	printf("numMissions: %d\n", sc->numMissions);
	printf("numExclusive: %d\n", sc->numExclusive);
	printf("mainOffset: 0x%x\n", sc->mainOffset);
	for(i = 0; i < sc->numMissions; i++)
		printf("miss %d: 0x%x\n", i, sc->missionOffsets[i]);
}

int
main(int argc, char *argv[])
{
	FILE *f;
	Script sc;
	long buflen;
	int i;
	char name[128];

	if(argc < 2){
		fprintf(stderr, "usage: %s file\n", argv[0]);
		return 1;
	}
	if(f = fopen(argv[1], "rb"), f == NULL){
		fprintf(stderr, "Couldn't open file: %s\n", argv[1]);
		return 1;
	}
	fseek(f, 0, SEEK_END);
	buflen = ftell(f);
	rewind(f);
	sc.buf = malloc(buflen);
	sc.bufSize = buflen;
	fread(sc.buf, 1, buflen, f);
	fclose(f);

	scriptInit(&sc);
	dumpScriptInfo(&sc);

	sprintf(name, "disasm/main.sc");
	if(f = fopen(name, "wb"), f == NULL){
		fprintf(stderr, "Couldn't open file: %s\n", name);
		return 1;
	}
	disassemble(f, &sc, sc.buf+sc.mainOffset, sc.buf+sc.missionOffsets[0]);
	fclose(f);

	for(i = 0; i < sc.numMissions-1; i++){
		sprintf(name, "disasm/%d.sc", i);
		if(f = fopen(name, "wb"), f == NULL){
			fprintf(stderr, "Couldn't open file: %s\n", name);
			return 1;
		}
		disassemble(f, &sc, sc.buf+sc.missionOffsets[i],
		            sc.buf+sc.missionOffsets[i+1]);
		fclose(f);
	}

	return 0;
}
